package limax.codec.asn1;

import java.io.ByteArrayOutputStream;
import java.nio.charset.StandardCharsets;
import java.util.Collection;

import limax.codec.Codec;
import limax.codec.CodecException;

public class ASN1IA5String extends ASN1Object implements PrimitiveObject, ConstructedObject {
	public static final ASN1ObjectIdentifier oid = new ASN1ObjectIdentifier("2.1.0.0");
	public static final ASN1Tag tag = new ASN1Tag(TagClass.Universal, 22);

	private final ASN1PrimitiveObject obj;

	public ASN1IA5String() {
		obj = new ASN1PrimitiveObject(tag);
	}

	public ASN1IA5String(ASN1Tag tag) {
		obj = new ASN1PrimitiveObject(tag);
	}

	public ASN1IA5String(ASN1Tag tag, String value) {
		obj = new ASN1PrimitiveObject(tag);
		setData(value.getBytes(StandardCharsets.ISO_8859_1));
	}

	public ASN1IA5String(String value) {
		this(tag, value);
	}

	public String get() {
		return new String(getData(), StandardCharsets.ISO_8859_1);
	}
	
	@Override
	public void setData(byte[] data) {
		obj.setData(data);
	}

	@Override
	public byte[] getData() {
		return obj.getData();
	}

	@Override
	public boolean isEmpty() {
		return obj.isEmpty();
	}

	@Override
	public long renderLength() {
		return obj.renderLength();
	}

	@Override
	public void render(Codec c) throws CodecException {
		obj.render(c);
	}

	@Override
	public String toString() {
		return "ASN1IA5String [" + new String(getData(), StandardCharsets.UTF_8) + "]";
	}

	@Override
	public ASN1ObjectIdentifier getOID() {
		return oid;
	}

	@Override
	public ASN1Tag getTag() {
		return obj.getTag();
	}

	@Override
	public void addChildren(Collection<ASN1Object> objects) {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		objects.stream().map(obj -> ((ASN1IA5String) obj).getData()).forEach(data -> out.write(data, 0, data.length));
		setData(out.toByteArray());
	}
}
